# ==================================================
# some nice things about this class
# ==================================================

import appdirs
import os
import yaml
import re
import copy

from izaber.compat import *
from izaber.startup import initializer, app_config
from izaber.structs import DictObj, deep_merge

class YAMLConfig(object):
    _app_name = 'ZaberConfig'
    _app_author = 'Zaber'
    _config_filename = 'izaber.yaml'
    _config_dirs = [
                    '.',
                    os.path.expanduser('~'),
                    appdirs.user_data_dir(_app_name, _app_author),
                  ]
    _overlays = []

    _cfg = None
    _cfg_merged = None

    # ================================================
    # constructor
    # ================================================
    def __init__(self,*args,**kwargs):
        if args or kwargs:
            self.load_config(*args,**kwargs)

    # ================================================
    # helpers
    # ================================================
    def __repr__(self):
        """ Returns the YAML representation of the configuration
        """
        return yaml.dump(self._cfg, default_flow_style=False)

    def __nonzero__(self):
        return self._cfg or False

    def dict(self):
        return self._cfg_merged

    def config_find(self,config_dirs=None,config_filename=None):
        """ Attempt to use the config dir/config_filenames to
            locate the configuration file requested. Some folks
            would prefer to keep their config in ~ where it's in
            plain sight rather than the buried application
            specific location
        """
        if config_dirs is None:
            config_dirs = self._config_dirs
        else:
            if isinstance(config_dirs,basestring):
                config_dirs = [config_dirs]

        if config_filename is None:
            config_filename = self._config_filename
        for test_dir in config_dirs:
            test_fpath = os.path.join(test_dir,config_filename)
            if os.path.isfile(test_fpath):
                return test_fpath

        # No matches found
        return

    def load_config( self,
                  config_filename=None,
                  config_dirs=None,
                  config_buffer=None,
                  config_amend=None,
                  environment=None,
                  overlays=None,
                ):
        # Does the actual work of loading

        # Setup defaults
        if config_filename:
            self._config_filename = config_filename
        if config_dirs:
            if isinstance(config_dirs,basestring):
                self._config_dirs = [config_dirs]
            else:
                self._config_dirs = config_dirs

        if config_buffer:
            self._config_full_filname = None
            self._cfg = yaml.safe_load(config_buffer)
        else:
            self.config_fpath = self.config_find() \
                                          or os.path.join(self._config_dirs[0], \
                                             self._config_filename)

            # check if config directory exists, and create if necessary
            self._config_dir = os.path.dirname(self.config_fpath)
            if not os.path.exists(self._config_dir):
                os.makedirs(self._config_dir)

            try:
                with open(self.config_fpath,'r') as file_obj:
                    self._cfg = yaml.full_load(file_obj)
            except IOError:
                self._cfg = {}

        # initialize environment
        if environment == None:
            environment = 'default'
        self._env = environment

        # Apply the overlay if required
        if not overlays:
            overlays = []
        self._overlays = overlays
        self.overlay_load()

        # Amend the configurations if required
        if config_amend:
            self.config_amend_(config_amend)

    def overlay_load(self):
        _cfg_merged = copy.deepcopy(self._cfg.get('default',{}))
        if self._env not in self._cfg:
            raise Exception("Could not find environment '%s' in configuration"%self._env)
        _cfg_merged = deep_merge(_cfg_merged,self._cfg[self._env])
        for overlay in self._overlays:
            _cfg_merged = deep_merge(_cfg_merged,overlay)
        self._cfg_merged = _cfg_merged

    def overlay_add(self,overlay):
        self._overlays.append(overlay)
        self._cfg_merged = deep_merge(self._cfg_merged,overlay)

    # ================================================
    # dealing with dynamic attributes
    # ================================================
    def get(self,k,default=None):
        # Try for a direct key match regardless of type.
        try:
            return self._cfg_merged[k]
        except KeyError:
            pass
        # Try for a dot-notation match.
        try:
            ref = self._cfg_merged
            for tok in k.split("."):
                ref = ref[tok]
            return ref
        except (KeyError, TypeError, AttributeError):
            # Finally, no match, so return default.
            return default

    def __getattr__(self, name):
        if name not in self._cfg_merged:
            raise AttributeError("{} is not an attribute".format(name))
        v = self._cfg_merged[name]
        if isinstance(v,dict):
            return DictObj(self,v)
        return v

    def __getitem__(self,name):
        return self.__getattr__(name)

    def addon_config(self,name):
        if name in self._cfg_merged:
            v = self._cfg_merged[name]
        else:
            v = {}
            self._cfg[self._env][name] = v
            self._cfg_merged[name] = v
        return DictObj(self,v)

    # ================================================
    # choose the environment
    # ================================================
    def environment(self, env=None):
        if env != None:
            self._env = env
            if self._env not in self._cfg:
                self._cfg[self._env] = {}
        return self._env

    # ================================================
    # Configuration Amending/Extending
    # ================================================
    def cfg_(self,cfg=None):
        """
        Getter/Setter of configuration data. This can be used
        to update and modify the configuration file on the system
        by new applications.
        """
        if cfg is None:
            cfg = self._cfg
        else:
            self._cfg = cfg
        self.overlay_load()
        return cfg

    def config_amend_key_(self,key,value):
        """ This will take a stringified key representation and value and
        load it into the configuration file for furthur usage. The good
        part about this method is that it doesn't clobber, only appends
        when keys are missing.
        """
        cfg_i = self._cfg
        keys = key.split('.')
        last_key = keys.pop()
        trail = []
        for e in keys:
            cfg_i.setdefault(e,{})
            cfg_i = cfg_i[e]
            trail.append(e)
            if not isinstance(cfg_i,dict):
                raise Exception('.'.join(trail) + ' has conflicting dict/scalar types!')
        cfg_i.setdefault(last_key,value)

    def config_amend_(self,config_amend):
        """ This will take a YAML or dict configuration and load it into
            the configuration file for furthur usage. The good part
            about this method is that it doesn't clobber, only appends
            when keys are missing.

            This should provide a value in dictionary format like:

            {
              'default': {
                'togglsync': {
                  'dsn': 'sqlite:///zerp-toggl.db',
                  'default': {
                    'username': 'abced',
                    'toggl_api_key': 'arfarfarf',
                  },
                  'dev': {
                    'cache': False
                  }
               }
            }

            OR at user's preference can also use yaml format:

            default:
              togglsync:
                  dsn: 'sqlite:///zerp-toggl.db'
                  default:
                    username: 'abced'
                    toggl_api_key: 'arfarfarf'
                  dev:
                    cache: False

            Then the code will append the key/values where they may be
            missing.

            If there is a conflict between a dict key and a value, this
            function will throw an exception.

            IMPORTANT: after making the change to the configuration, 
                       remember to save the changes with cfg.save_()

        """
        if not isinstance(config_amend,dict):
            config_amend = yaml.full_load(config_amend)

        def merge_dicts(source,target,breadcrumbs=None):
            """
            Function to update the configuration if required. Returns
            True if a change was made.
            """

            changed = False

            if breadcrumbs is None:
                breadcrumbs = []

            # Don't descend if we're not a dict
            if not isinstance(source,dict):
              return source

            # Let's start iterating over things
            for k,v in source.items():

                # New key, simply add.
                if k not in target:
                    target[k] = v
                    changed = True
                    continue

                # Not new key.... so is it a dict?
                elif isinstance(target[k],dict):
                    trail = breadcrumbs+[k]
                    if isinstance(v,dict):
                        if merge_dicts(v,target[k],trail):
                            changed = True
                    else:
                        raise Exception('.'.join(trail) + ' has conflicting dict/scalar types!')

                else:
                    trail = breadcrumbs+[k]
                    if isinstance(v,dict):
                        raise Exception('.'.join(trail) + ' has conflicting dict/scalar types!')

            return changed

        if merge_dicts(config_amend,self._cfg):
            self.overlay_load()

        return self._cfg


    def config_update_key_(self,key,value):
        """ This will take a stringified key representation and value and
        load it into the configuration file for furthur usage.
        This method will clobber existing entries allowing something like a
        "reset"
        """
        cfg_i = self._cfg
        keys = key.split('.')
        last_key = keys.pop()
        trail = []
        for e in keys:
            cfg_i.setdefault(e,{})
            cfg_i = cfg_i[e]
            trail.append(e)
            if not isinstance(cfg_i,dict):
                raise Exception('.'.join(trail) + ' has conflicting dict/scalar types!')
        cfg_i[last_key] = value

    def config_update_(self,config_amend):
        """ This will take a YAML or dict configuration and load it into
            the configuration file for furthur usage.
            This method will clobber existing entries allowing something like a
            "reset"

            This should provide a value in dictionary format like:

            {
              'default': {
                'togglsync': {
                  'dsn': 'sqlite:///zerp-toggl.db',
                  'default': {
                    'username': 'abced',
                    'toggl_api_key': 'arfarfarf',
                  },
                  'dev': {
                    'cache': False
                  }
               }
            }

            OR at user's preference can also use yaml format:

            default:
              togglsync:
                  dsn: 'sqlite:///zerp-toggl.db'
                  default:
                    username: 'abced'
                    toggl_api_key: 'arfarfarf'
                  dev:
                    cache: False

            Then the code will append the key/values where they may be
            missing.

            If there is a conflict between a dict key and a value, this
            function will throw an exception.

            IMPORTANT: after making the change to the configuration, 
                       remember to save the changes with cfg.save_()

        """
        if not isinstance(config_amend,dict):
            config_amend = yaml.safe_load(config_amend)

        def merge_dicts(source,target,breadcrumbs=None):
            """
            Function to update the configuration if required. Returns
            True if a change was made.
            """

            changed = False

            if breadcrumbs is None:
                breadcrumbs = []

            # Don't descend if we're not a dict
            if not isinstance(source,dict):
              return source

            # Let's start iterating over things
            for k,v in source.items():

                # New key, simply add.
                if k not in target:
                    target[k] = v
                    changed = True
                    continue

                # Not new key.... so is it a dict?
                elif isinstance(target[k],dict):
                    trail = breadcrumbs+[k]
                    if isinstance(v,dict):
                        if merge_dicts(v,target[k],trail):
                            changed = True
                    else:
                        raise Exception('.'.join(trail) + ' has conflicting dict/scalar types!')

                else:
                    trail = breadcrumbs+[k]
                    if isinstance(v,dict):
                        raise Exception('.'.join(trail) + ' has conflicting dict/scalar types!')

            return changed

        if merge_dicts(config_amend,self._cfg):
            self.overlay_load()

        return self._cfg


    # ================================================
    # write config to yaml file
    # ================================================
    def save_(self):
        if self.config_fpath == None:
            raise Exception("Cannot save config when ")
        file_obj = open(self.config_fpath, 'w')
        yaml.dump(self._cfg, file_obj, default_flow_style=False)
        file_obj.close()
        return self.config_fpath

# Global shared YAML configuration
config = YAMLConfig()

@initializer('config')
def initialize(**kwargs):
    """
    Loads the globally shared YAML configuration
    """
    global config
    config_opts = kwargs.setdefault('config',{})

    if isinstance(config_opts,basestring):
        config_opts = {'config_filename':config_opts}
        kwargs['config'] = config_opts

    if 'environment' in kwargs:
        config_opts['environment'] = kwargs['environment']

    config.load_config(**config_opts)

    # Overlay the subconfig
    if kwargs.get('name'):
        subconfig = config.get(kwargs.get('name'),{})
        config.overlay_add(subconfig)

    config.overlay_add(app_config)




# -*- coding: utf-8 -*-

import os
import re
import datetime
import logging

import jinja2

from izaber.compat import *

from izaber.startup import initializer, request_initialize, app_config
from izaber import config

def parse(template,**tags):
    now = datetime.datetime.now()
    tags.update({
        'date': now.strftime('%Y-%m-%d'),
        'time': now.strftime('%H:%M:%S'),
        'dt': now.strftime('%Y-%m-%d_%H-%M-%S'),
    })
    path = jinja2.Template(template).render(**tags)
    return os.path.expanduser(path)

class Path(object):
    path_template = None
    tags = None
    path = None

    def __init__(self,path_template,**tags):
        if isinstance(path_template,Path):
            self.path_template = path_template.path_template
        else:
            self.path_template = path_template
        self.tags = tags
        self.path = self.parse(self.path_template or '.')

    def parse(self,buf,**tags):
        tags.update(self.tags)
        return parse(buf,**tags)

    def file_fpath(self,fpath,**tags):
        parsed_fpath = self.parse(fpath,**tags)
        full_fpath = os.path.join(unicode(self.path),parsed_fpath)
        return full_fpath

    def open(self,fpath,*args,**kwargs):
        return File(self.path_template,fpath).open(*args,**kwargs)

    def exists(self):
        return os.path.exists(self.path)

    def makedirs(self):
        return os.makedirs(self.path)

    def __str__(self):
        return self.path

    def __unicode__(self):
        return self.path

    def walk(self,visit,arg):
        os.path.walk(self.path,visit,arg)

class File(Path):
    full_fpath = ''
    file_handle = None

    def __init__(self,path_template,fpath,**tags):
        super(File,self).__init__(path_template,**tags)
        self.full_fpath = self.file_fpath(fpath)

    def __str__(self):
        return self.full_fpath

    def __unicode__(self):
        return self.full_fpath

    def open(self,*args,**kwargs):
        self.file_handle = open(self.full_fpath,*args,**kwargs)
        return self

    def __getattr__(self,k):
        return getattr(self.file_handle,k)

class DataDir(object):

    def __init__(self,*args,**kwargs):
        if args or kwargs:
            self.initialize(*args, **kwargs)

    def initialize(self,path=None,**kwargs):

        # Get the initial information
        path_templates = {}
        tags = {'path':path}
        for k,v in kwargs.items():
            if re.search('_path$',k):
                path_templates[k] = v
            else:
                tags[k] = v

        # Then get the base path
        self.path = Path(path,**kwargs)

        resolved_paths = {
            'path': self.path
        }
        seen = {}
        class SneakyString(object):
            def __init__(self,path_key,path_template):
                self.path_key = path_key
                self.path_template = path_template

            def path_tag(self,k,path_template,seen=None):
                if not seen: seen = {}
                if k in seen:
                    raise Exception('Cyclic dependancy found')
                if not k in resolved_paths:
                    full_path = parse(path_template,**tags)
                    resolved_paths[k] = Path(full_path,**tags)
                return resolved_paths[k]

            def __unicode__(self):
                return unicode(self.path_tag(self.path_key,self.path_template))

            def __str__(self):
                return str(self.path_tag(self.path_key,self.path_template))

        for k,path_template in path_templates.items():
            tags[k] = SneakyString(k,path_template)

        for k,path_template in path_templates.items():
            if k in resolved_paths: continue
            full_path = parse(path_template,**tags)
            resolved_paths[k] = Path(full_path,**tags)

        self.path_templates = path_templates
        self.tags = tags
        self.paths = resolved_paths

        # If any of the data directories need to be made, 
        # go ahead and do so
        for k,path in self.paths.items():
            if not path.exists():
                path.makedirs()

    def full_fpath(self,fpath,**kwargs):
        tags = dict(self.tags)
        tags.update(self.paths)
        tags.update(kwargs)
        return parse(fpath,**tags)

    def file_get(self, fname,**kwargs):
        return File(
                    fpath=fname,
                    path_template=self.path,
                    **kwargs
                )

    def path_get(self, path_template, **kwargs):
        return Path(
                    path_template=path_template,
                    **kwargs
                )

    def logger(self,fpath=None,log_level=None):
        """ Creates a custom logger
        """
        log = logging.getLogger(fpath)
        pass

    def open(self,fname,mode=None,*args,**kwargs):
        f = self.file_get(fname,**kwargs).open(mode=mode)
        return f

    def __getattr__(self,k):
        if self.paths and k in self.paths:
            return self.paths[k]
        raise AttributeError("{} is not found in {}".format(k,self.paths))

paths = DataDir()

logger_defaults = {
  'filemode': 'a',
  # 'format': '%(levelname)s %(asctime)s %(module)s: %(message)s',
  'datefmt': '%Y-%m-%d %H:%M:%S',
  'level': logging.WARNING
}

class Logger(logging.Logger):

    _iz_filename = None
    _iz_handler = None
    _iz_format = None
    _iz_name = None

    def __init__(self,name=None,**kwargs):
        super(Logger,self).__init__(name)
        self._iz_name = name
        if kwargs:
            self.init(**kwargs)

    def init(self,**kwargs):

        # Load the gloabl config if any
        log_config = dict(logger_defaults)
        log_config.update(config.get('log',{}))
        log_config.update(app_config.get('log',{}))
        log_config.update(config.get('log',{}).get(self._iz_name,{}))
        log_config.update(app_config.get('log',{}).get(self._iz_name,{}))
        log_config.update(kwargs)

        if log_config.get('level') and self._iz_handler:
            self._iz_handler.setLevel(log_config['level'])

        if log_config.get('filename') and not self._iz_filename:
            filename = paths.full_fpath(log_config['filename'])
            self._iz_handler = logging.FileHandler(
                                    filename=filename,
                                    mode=log_config.get('filemode','a'),
                                    encoding=log_config.get('fileencoding','utf8'),
                                )
            self._iz_handler.setLevel(logging.DEBUG)
            self.addHandler(self._iz_handler)

        if log_config.get('formatter') and self._iz_handler:
            self._iz_format = log_config['formatter']
            self._iz_handler.setFormatter(log_config['formatter'])

        if log_config.get('format') and not self._iz_format:
            formatter = logging.Formatter(
                fmt=log_config['format'],
                datefmt=log_config.get('dateformat'),
            )
            self._iz_handler.setFormatter(formatter)
            self._iz_format = log_config.get('format')

        if log_config.get('level') and self._iz_handler:
            self._iz_handler.setLevel(log_config['level'])

logging.setLoggerClass(Logger)

log = logging.getLogger()

def getLogger(name,**kwargs):
    log = logging.getLogger(name)
    log.init(**kwargs)
    return log

@initializer('paths')
def initialize(**kwargs):

    request_initialize('config',**kwargs)

    # Setup the paths
    # Path configuration should be in the format:
    # paths: {
    #     path: path to main data directory
    #
    #     custom_path1: encoded path
    #     ...
    #     custom_pathN: encoded path
    #
    # }

    # Setup the paths first
    paths_config = config.get('paths',{})
    paths_config.update(kwargs.get('paths',{}))
    if 'path' not in paths_config:
        for path_location in ['/tmp','/temp','.']:
            if os.path.isdir(path_location):
                paths_config.setdefault('path',path_location)
                break
    paths.initialize(**paths_config)

    # Setup the logger
    # Log configuration should be in the format:
    # log: {
    #     filename: paths compatible filepath
    #     filemode: usually 'a'
    #     fileencoding: usually 'utf8'
    #     format: logs formatter compatible format
    #     dateformat: how to display dates
    #     level: filter out everything above this level
    #
    #     custom_logger_name1: {
    #         filename: paths compatible filepath
    #         filemode: usually 'a'
    #         fileencoding: usually 'utf8'
    #         format: logs formatter compatible format
    #         dateformat: how to display dates
    #         level: filter out everything above this level
    #     },
    #     ...
    #     custom_logger_nameN: {
    #         filename: paths compatible filepath
    #         filemode: usually 'a'
    #         fileencoding: usually 'utf8'
    #         format: logs formatter compatible format
    #         dateformat: how to display dates
    #         level: filter out everything above this level
    #     },
    # }
    #
    log_config = dict(logger_defaults)
    log_config.update(config.get('log',{}))
    log_config.update(kwargs.get('log',{}))

    # Default to app_dir/izaber.log if path not defined
    filename = paths.file_get('izaber.log')
    log_config.setdefault('filename',unicode(filename))
    log_config['filename'] = paths.full_fpath(log_config['filename'])

    logging.basicConfig(**log_config)
    logging.getLogger().addHandler(logging.StreamHandler())




from __future__ import absolute_import

import os.path
import smtplib
import logging

from email.parser import Parser
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.base import MIMEBase
from email import encoders

from bs4 import BeautifulSoup

from izaber import config, app_config
from izaber.startup import request_initialize, initializer
from izaber.compat import *
from izaber.paths import paths
from izaber.templates import parse, parsestr

log = logging.getLogger('email')

class Mailer(object):
    def load_config(self,**options):
        self.options = options

    def sendmail(self,**kwargs):
        if config.get('debug') or self.options.get('debug'):
            msg = kwargs['msg']
            log.info(
                'DID NOT send email "{}" from "{}" to {}'.format(
                    msg['Subject'],
                    kwargs.get('from_addr'),
                    msg['To'],
                )
            )
            log.debug(msg.as_string())
            return

        log.info(
            'Sent email "{}" from "{}" to {}'.format(
                kwargs['msg']['Subject'],
                kwargs.get('from_addr'),
                kwargs['msg']['To'],
            )
        )
        host = self.options.get('host',config.email.host)
        server = smtplib.SMTP(host)
        kwargs['msg'] = kwargs['msg'].as_string()
        server.sendmail(**kwargs)
        server.quit()

    def attachment_create(self,fpath):
        full_fpath = paths.full_fpath(fpath)
        with open(full_fpath,'rb') as f:
            # FIXME: Do I need to handle mimetype?
            att = MIMEBase('application', "octet-stream")
            att.set_payload(f.read())
            encoders.encode_base64(att)
        att.add_header(
            'Content-Disposition',
            'attachment; filename="{}"'.format(os.path.basename(fpath))
        )
        return att

    def message_send(self,msg):
        return self.sendmail(
                    from_addr=msg['from'],
                    to_addrs=msg['to'].split(','),
                    msg=msg
                )

    def basic_send(
              self,
              from_addr,
              to_addrs,
              subject,
              body
          ):
        msg = MIMEText(body)
        msg['Subject'] = subject
        msg['From'] = from_addr
        msg['To'] = ", ".join(to_addrs)
        self.sendmail(
            from_addr=from_addr,
            to_addrs=to_addrs,
            msg=msg
        )

    def template_send(self,fpath,**kwargs):
        msg = self.template_parse(fpath,**kwargs)
        to = msg['to']
        if isinstance(to,basestring):
            to = [to.split(',')]
        self.sendmail(
            from_addr=msg['from'],
            to_addrs=to,
            msg=msg,
        )

    def template_parsestr(self,template,**tags):
        tags['config'] = config.dict()
        parsed_email = parsestr(template,**tags)
        if config.get('debug') or self.options.get('debug'):
            log.debug(u"---------\n"+parsed_email+"\n--------\n")
        return self.message_fromstr(parsed_email)

    def template_parse(self,fpath,**tags):
        tags['config'] = config.dict()
        parsed_email = parse(fpath,**tags)
        if config.get('debug') or self.options.get('debug'):
            log.debug(u"---------\n"+parsed_email+"\n--------\n")
        return self.message_fromstr(parsed_email)

    def template_sendstr(self,template,**kwargs):
        msg = self.template_parsestr(template,**kwargs)
        to = msg['to']
        if isinstance(to,basestring):
            to = [to.split(',')]
        self.sendmail(
            from_addr=msg['from'],
            to_addrs=to,
            msg=msg
        )

    def message_fromstr(self,parsed_email):
        # Decompose the email into constituent parts first
        elements = parsed_email.split('\n')
        headers = []
        content = []
        state = 'headers'
        for e in elements:
            if state == 'headers':
                if not e:
                    state = 'content'
                headers.append(e)
            else:
                content.append(e)

        # Parse the headers
        headers = u"\n".join(headers)+u"\n\n"
        e = Parser().parsestr(headers)

        # Reconstruct the email into mime formatted elements.
        msg = MIMEMultipart('mixed')
        for k,v in dict(e).items():
            msg[k] = v
        msg_text = MIMEMultipart('alternative')
        html = u"\n".join(content)
        text = BeautifulSoup(html,"lxml").get_text()
        msg_text.attach(MIMEText(text,'plain',_charset='utf-8'))
        msg_text.attach(MIMEText(html,'html',_charset='utf-8'))
        msg.attach(msg_text)
        return msg

mailer = Mailer()

@initializer('email')
def load_config(**options):
    request_initialize('config',**options)
    email_config = config.email.dict()
    mailer.load_config(**email_config)




# Adds support for various templates
import jinja2
import datetime

from izaber.startup import initializer, request_initialize, app_config
from izaber.paths import paths
from izaber import config

compiled_templates = {}

template_loader = jinja2.FileSystemLoader(searchpath=[".","/"])
template_env = jinja2.Environment(loader=template_loader)

def tags_common():
    now = datetime.datetime.now()
    return {
        'now': now,
        'date': now.strftime('%Y-%m-%d'),
        'date_iso': now.isoformat(),
        'time': now.strftime('%H:%M:%S'),
        'dt': now.strftime('%Y-%m-%d_%H-%M-%S'),
        'config': config,
    }

def parse(template,**tags):
    template_fpath = paths.full_fpath(template)
    tags.update(tags_common())
    if template_fpath not in compiled_templates:
        compiled_templates[template_fpath] = \
            template_env.get_template(template_fpath)
    return compiled_templates[template_fpath].render(**tags)

def parsestr(template,**tags):
    now = datetime.datetime.now()
    tags.update(tags_common())
    return jinja2.Template(template).render(**tags)

@initializer('templates')
def initialize(**kwargs):
    request_initialize('config',**kwargs)




import logging
import os
import sys

from izaber.startup import initializer, request_initialize
from izaber.zconfig import config

class Logger(object):
    _logger = None

    def __call__(self,*args,**kwargs):
        self.info(*args,**kwargs)

    def __getattr__(self,k):
        if not self._logger:
            self._logger = logging.getLogger('main')
        return getattr(self._logger,k)

log = Logger()

@initializer('logging')
def initialize(**kwargs):
    request_initialize('config',**kwargs)
    logging_config = config.get('logging',{})
    logging_config.update( kwargs.get('logging',{}) )

    # We only activate logging if the logging section has
    # configuration. This allows users to configure their
    # own logger as desired
    if not logging_config:
        return

    # We also allow the disabling of the internally made logging module
    # by adding the logging.disable_internal key
    if logging_config.get('disable_internal'):
        return

    # So we have some logging configuration, let's let izaber setup a logging
    # handler.

    # If there's no filename, let's default to stream output for logging
    if 'filename' not in logging_config:
        logging_config.setdefault('stream', sys.stdout)

    # so there's a filename but it's blank, let's default to something
    elif not logging_config['filename']:
        for log_location in ['/tmp','/temp','.']:
            if not os.path.isdir(log_location):
                continue
            logging_config.setdefault('filename',os.path.join(log_location,'log.log'))
            break
    logging_config.setdefault('filemode','a')
    logging_config.setdefault(
        'format',
        '%(levelname)s %(asctime)s %(module)s: %(message)s'
    )
    logging_config.setdefault('datefmt','%Y-%m-%d %H:%M:%S')
    logging_config.setdefault('level',logging.WARNING)

    logging.basicConfig(**logging_config)




import sys
import re
import importlib

try:
    from importlib.metadata import version
    __version__ = version("izaber")
# For older versions of python
except:
    import pkg_resources
    __version__ = pkg_resources.get_distribution("izaber").version

# Check to see the library has the submodule we need
class IZaberLoaderImportlib(object):

    def __init__(self,spec,module_name,*args,**kwargs):
        self.spec = spec
        self.module_name = module_name
        self.args = args
        self.kwargs = kwargs

    def load_module(self, module_name):
        # In the case of pyinstaller, it seems to set sys.frozen
        # we check for that and load the module directly
        if hasattr(sys,'frozen'):
            module = self.spec.loader.load_module(self.spec.name)
            sys.modules[self.module_name] = module
            sys.modules[module_name] = module
            sys.modules[self.spec.name] = module
            self.spec.loader.exec_module(module)
        else:
            if self.spec.name in sys.modules:
                module = sys.modules[self.spec.name]
                sys.modules.setdefault(self.module_name, module)
                sys.modules.setdefault(module_name, module)
            else:
                module = importlib.util.module_from_spec(self.spec)
                sys.modules[self.module_name] = module
                sys.modules[module_name] = module
                sys.modules[self.spec.name] = module
                self.spec.loader.exec_module(module)
        return module

class IZaberSpec:
    def __init__(self, spec, module_name):
        self.spec = spec
        self.module_name = module_name
        self.loader = IZaberLoaderImportlib(spec, module_name)

    def __getattr__(self, k):
        return getattr(self.spec, k)

from importlib import abc as il_abc
class IZaberFinderImportlib(il_abc.MetaPathFinder):
    """ Attempts to finds the module that may be tucked into
        a submodule.
    """

    def __init__(self,prefixes=None):
        super(IZaberFinderImportlib,self).__init__()
        if prefixes is None:
            prefixes = ['izaber.']
        self.prefixes = prefixes

    def add_prefix(self,prefix):
        normalized_prefix = prefix+'.'
        if normalized_prefix not in self.prefixes:
            self.prefixes.append(normalized_prefix)
            self.prefixes.sort(key=lambda a:len(a),reverse=True)

    def attempt_load(self,prefix,module_name):
        # Extension modules are expected to be named:
        #
        # izaber_extensionname
        #
        # where extensionname is all alphabetical 
        try:
            target_module = re.sub(
                          '^'+prefix,
                          prefix.replace('.','_'),
                          module_name,
                      )

            found = None
            spec = None
            package_path = []
            for e in target_module.split('.'):
                package = "_".join(package_path) or None
                spec = importlib.util.find_spec(e,package)
                if not spec:
                    raise ModuleNotFoundError(f"No module named {module_name}")
                package_path.append(e)

            return IZaberLoaderImportlib(spec,module_name)
        except ImportError:
            return

    def find_spec(self, module_name, path=None, target=None):

        for prefix in self.prefixes:
            try:
                if module_name.index(prefix) != 0:
                    continue
            except ValueError:
                continue

            try:
                target_module = re.sub(
                              '^'+prefix,
                              prefix.replace('.','_'),
                              module_name,
                          )

                found = None
                spec = None
                package_path = []
                for e in target_module.split('.'):
                    package = "_".join(package_path) or None
                    spec = importlib.util.find_spec(e,package)
                    if not spec:
                        raise ModuleNotFoundError(f"No module named {module_name}")
                    package_path.append(e)

                return IZaberSpec(spec,module_name)
            except ImportError:
                return

        return

    def find_module(self, module_name, package_path=None):
        # Only respond to izaber.* modules
        try:
            if module_name.index('izaber.') != 0:
                return
        except:
            return

        for prefix in self.prefixes:
            try:
                if module_name.index(prefix) != 0:
                    continue
            except ValueError:
                continue
            loader = self.attempt_load(prefix,module_name)
            if loader:
                return loader

        # Couldn't find it!
        return

import traceback

from izaber.startup import initialize, app_config
from izaber.zconfig import config

"""
This module does some magic! Without the following code it would
not be able to share the izaber.* namespace.

What we do here is intercept the request for any "import"
statement that starts with "izaber". If it does, it attempts to
rewrite the import attempt to izaber_WHATEVERTHENEXTELEMENTWAS

Then if it manages to find and load it okay, it will then pass
the loaded import onto to the rest of the system under the
izaber.* namespace for usage by developers.

If a submodule should also serve as the base point for additional
submodules, be sure to include a call to add_prefix. For instance,
if izaber.foo should allow the loading of izber.foo.bar via
izaber_foo_bar, then ensure that izaber_foo/__init__.py contains
the following code:

import izaber
izaber.autoloader.add_prefix('izaber_foo')

For another example have a look at izaber.wamp (which allows the
load of izaber.wamp.fuse)

"""


autoloader = IZaberFinderImportlib()
sys.meta_path.insert(0, autoloader)






def deep_merge(base,overlay):

    # If we have reached the end of either of the
    # base or the overlay we are done
    if overlay is None: return base
    if base is None: return overlay

    # Ensure the types are the same
    if not isinstance(base,type(overlay)):
        raise Exception('Data structures cannot be combined.')

    if isinstance(base,dict):
        for k,v in overlay.items():
            base[k] = deep_merge(base.get(k),v)
        return base

    if isinstance(base,list):
        # FIXME: how to deal with lists?
        # Do we add the lists together
        return overlay
    else:
        return overlay

class DictObj(object):
    def __init__(self,config,data):
        self._config = config
        self._data = data

    def dict(self):
        return self._data

    def get(self,*args,**kwargs):
        return self._data.get(*args,**kwargs)

    def __getattr__(self,k):
        v = self._data[k]
        if isinstance(v,dict):
            return DictObj(self._config,v)
        return v

    def __getitem__(self,k):
        return self.__getattr__(k)

    def __str__(self):
        return str(self._data)

    def __call__(self):
        return self._data




# Stolen from http://www.rfk.id.au/blog/entry/preparing-pyenchant-for-python-3/

try:
    unicode = unicode
except NameError:
    # 'unicode' is undefined, must be Python 3
    str = str
    unicode = str
    bytes = bytes
    basestring = (str,bytes)
else:
    # 'unicode' exists, must be Python 2
    str = str
    unicode = unicode
    bytes = str
    basestring = basestring





import pytz
import six
import dateutil.parser
import re
import datetime

###########################################
# Time and Date handlers
###########################################

def time_range_cutter_at_time(local,time_range,time_cut=(0,0,0)):
    """ Given a range, return a list of DateTimes that match the time_cut
        between start and end.

        :param local: if False [default] use UTC datetime. If True use localtz
        :param time_range: the TimeRange object
        :param time_cut: HH:MM:SS of when to cut. eg: (0,0,0) for midnight
    """

    ( start, end ) = time_range.get(local)
    index = start.replace(
                        hour=time_cut[0],
                        minute=time_cut[1],
                        second=time_cut[2]
                    )
    cuts = []
    index += datetime.timedelta(days=1)
    while index < end:
        cuts.append(index)

        index += datetime.timedelta(days=1)
        if local:
            index = time_range.normalize(index)
    return cuts

class TimeRange(object):
    start_time = None
    end_time = None

    def __init__(self, start_time, end_time):
        self.start_time = start_time
        self.end_time = end_time

    def normalize(self,dt):
        return self.start_time.local_tz.normalize(dt)


    def get(self,local=False):
        return (
            self.start_time.get(local),
            self.end_time.get(local),
        )

    def delta(self,local=False):
        """ Returns the number of days of difference
        """
        (s,e) = self.get(local)
        return e-s

    def hours(self,local=False):
        """ Returns the number of hours of difference
        """
        delta = self.delta(local)
        return delta.total_seconds()/3600


    def days(self,local=False):
        """ Returns the number of days of difference
        """
        delta = self.delta(local)
        return delta.days

    def __contains__(self,timestamp):
        """ Returns true if the timestamp is found within this object's range

            :param timestamp: DateTime object of the time to test
        """
        if isinstance(timestamp,DateTime):
            return (
                    self.start_time >= timestamp
                and self.end_time <= timestamp
            )
        # TODO: also compare against another TimeRange in the future?
        else:
            return False

    def chunks(self,local,cutter_callback=None,*cutter_callback_args,**cutter_callback_kwargs):
        """ Takes a time range and returns sub timeranges based upon
            the cuts that cutter callback provides

            :param local: if False [default] use UTC datetime. If True use localtz
            :param cutter_callback: This should be a callback function that
                    takes the current TimeRange and returns a list of
                    DateTimes that denote where the next chunk should
                    start. By default the cutter used will cut at 00:00:00
                    each day.
        """

        # First we get all the slices we want to take out of this time range.
        if not cutter_callback:
            cutter_callback = time_range_cutter_at_time
        time_cuts = cutter_callback(local,self,*cutter_callback_args,**cutter_callback_kwargs)

        # Now we need to make the time range objects for the cuts
        time_chunks = []
        time_index = self.start_time
        time_cuts = sorted(time_cuts)
        for time_cut in time_cuts:

            # FIXME: Better error message is probably going to be helpful
            if self.end_time < time_cut or self.start_time > time_cut:
                raise Exception('Cut time provided that is outside of time range')

            # Create the new tail end entry for times, combine it with the
            # index to craft a timerange and add it to the chunk list
            cut_end_time = time_cut

            # If the chunk is not the final chunk, we want to pull
            # the time back by a small smidgen
            if cut_end_time != self.end_time:
                cut_end_time -= datetime.timedelta(microseconds=1)

            time_ending = DateTime(
                            data=cut_end_time,
                            data_tz=time_index.local_tz,
                            local_tz=time_index.local_tz
                            )


            chunk_range = TimeRange(time_index,time_ending)
            time_chunks.append(chunk_range)

            # Setup the index for the next cut (the current cut
            # becomes the start of the next cut)
            time_index = DateTime(
                            data=time_cut,
                            data_tz=time_index.local_tz,
                            local_tz=time_index.local_tz
                        )

        # Add the last segment if required
        if time_index != self.end_time:
            time_chunks.append(
                TimeRange(time_index,self.end_time)
            )

        return time_chunks

    def str(self,local):
        """ Return the string representation of the time range

            :param local: if False [default] use UTC datetime. If True use localtz
        """
        s = self.start_time.str(local) \
            + u" to " \
            + self.end_time.str(local)
        return s

    def __getattr__(self,k):
        m = re.search('^(utc|local)_(.*)$',k)
        if m:
            local = m.group(1) == 'local'
            n = m.group(2)
            return lambda *a,**kw: getattr(self,n)(local,*a,**kw)

        return object.__getattr__(self,k)

class Date(object):

    local_tz = None
    data = None

    def __init__(self,data,local_tz=None,context=None):
        if isinstance(data,Date):
            self.local_tz = data.local_tz
            self.data = data.data
        else:
            if local_tz:
                self.local_tz = local_tz
            elif context and context.get('tz'):
                self.local_tz = pytz.timezone(context['tz'])
            else:
                self.local_tz = pytz.utc

            if isinstance(data,six.string_types):
                self.data = dateutil.parser.parse(data).date()
            else:
                self.data = data

    def localize(self,local=False,time_data=None):
        if not time_data: return time_data
        if local:
            return self.local_tz.localize(time_data)
        return pytz.utc.localize(time_data)

    @classmethod
    def now(cls,*args,**kwargs):
        date = datetime.datetime.now().date()
        return cls(date,*args,**kwargs)

    def start_time(self,local=False):
        time_data = datetime.datetime(
                            self.data.year,
                            self.data.month,
                            self.data.day,
                            0,
                            0,
                            0,
                            0,
                        )
        time_data = self.localize(local,time_data)
        return DateTime(time_data,self.local_tz,self.local_tz)

    def end_time(self,local=False):
        time_data = datetime.datetime(
                            self.data.year,
                            self.data.month,
                            self.data.day,
                            23,
                            59,
                            59,
                            999999,
                        )
        time_data = self.localize(local,time_data)
        return DateTime(time_data,self.local_tz,self.local_tz)

    def str(self,local=False):
        return self.data.isoformat()
    def __str__(self):
        return self.data.isoformat()

    def __getattr__(self,k):
        m = re.search('^(utc|local)_(.*)$',k)
        if m:
            local = m.group(1) == 'local'
            n = m.group(2)
            return lambda *a,**kw: getattr(self,n)(local,*a,**kw)

        return object.__getattribute__(self.data,k)

    def __add__(self,other):
        if isinstance(other,Date):
            return self.data + other.data
        else:
            return self.data + other

    def __eq__(self,other):
        if isinstance(other,Date):
            return self.data == other.data
        else:
            return self.data == other

    def __lt__(self,other):
        if isinstance(other,Date):
            return self.data < other.data
        else:
            return self.data < other

    def __le__(self,other):
        if isinstance(other,Date):
            return self.data <= other.data
        else:
            return self.data <= other

    def __gt__(self,other):
        if isinstance(other,Date):
            return self.data > other.data
        else:
            return self.data > other

    def __ge__(self,other):
        if isinstance(other,Date):
            return self.data >= other.data
        else:
            return self.data >= other

    def __add__(self,other):
        return self.data + Date(other).data

    def __sub__(self,other):
        return self.data - Date(other).data

class DateTime(object):

    # The timestamp we care about in UTC
    utc_data = None

    # The timezone we use for local transformations
    local_tz = None

    # The timestamp in local_tz
    local_data = None

    def __init__(self,data,data_tz,local_tz,context=None):
        """ Accepts either a string or an existing datetime and loads it up
        """
        self.local_tz = local_tz

        # Is this a string?
        if isinstance(data,six.string_types):
            timestamp = dateutil.parser.parse(data)

            # If we notice that there is an offset and that it's
            # UTC, let's ensure that it's using pytz.utc. There's this
            # weird error that I never got to the bottom of that somtimes
            # the timestamp.tzinfo turns into tzlocal. SO WEIRD! While
            # this isn't a true fix, as I don't know why it would randomly
            # switch to tzlocal despite a +00:00, this will let the code
            # do its job and "corrects" behaviour.
            if re.search(r'\+00:00',data):
                timestamp = timestamp.replace(tzinfo = pytz.utc)

            # Handle things differently based upon whether or not we have a timezone
            if timestamp.tzinfo:
                self.utc_data = timestamp.astimezone(pytz.utc)
                self.local_data = self.utc_data.astimezone(local_tz)
            else:
                timestamp = data_tz.localize(timestamp)
                self.utc_data = timestamp.astimezone(pytz.utc)
                self.local_data = timestamp.astimezone(local_tz)

        # Is it an existing DateTime?
        elif isinstance(data,DateTime):
            self.utc_data = data.utc_data
            self.local_tz = data.local_tz
            self.local_data = data.local_data

        # Is it a datetime?
        elif isinstance(data,datetime.datetime):
            # is it a naive datetime? if so we'll append a TZ to it
            if not data.tzinfo:
                timestamp = data_tz.localize(data)
                self.utc_data = timestamp.astimezone(pytz.utc)
                self.local_data = timestamp.astimezone(local_tz)

            else:
                self.utc_data = data.astimezone(pytz.utc)
                self.local_data = self.utc_data.astimezone(local_tz)

    def daily_hours(self,local=False):
        """ This returns a number from 0 to 24 that describes the number
            of hours passed in a day. This is very useful for hr.attendances
        """
        data = self.get(local)
        daily_hours = (data.hour +
                       data.minute / 60.0 +
                       data.second / 3600.0)
        return round(daily_hours,2)

    def get(self,local=False):
        """ Return the datetime object based upon whether we want the local
            or utc one.

            :param local: if False [default] return UTC datetime. If True return localtz
        """
        if local:
            return self.local_data
        return self.utc_data

    def date(self,local=False):
        """ Return the date object associated
            :param local: if False [default] return UTC date. If True return localtz date
        """
        return Date(self.get(local).date(),self.local_tz)

    def str(self,local=False,ifempty=None):
        """ Returns the string representation of the datetime
        """
        ts = self.get(local)
        if not ts: return ifempty
        return ts.strftime('%Y-%m-%d %H:%M:%S')

    def __getattr__(self,k):
        m = re.search('^(utc|local)_(.*)$',k)
        if m:
            local = m.group(1) == 'local'
            n = m.group(2)
            return lambda *a,**kw: getattr(self,n)(local,*a,**kw)

        return object.__getattr__(self,k)

    def __eq__(self,v):
        """ Handle == comparison
        """
        if isinstance(v,DateTime):
            return self.utc_data == v.utc_data
        elif isinstance(v,datetime.datetime):
            return self.utc_data == v
        return NotImplemented

    def __ne__(self,v):
        """ Handle != comparison
        """
        if isinstance(v,DateTime):
            return self.utc_data != v.utc_data
        elif isinstance(v,datetime.datetime):
            return self.utc_data != v
        return NotImplemented

    def __gt__(self,v):
        """ Handle > comparison
        """
        if isinstance(v,DateTime):
            return self.utc_data > v.utc_data
        elif isinstance(v,datetime.datetime):
            return self.utc_data > v
        return NotImplemented

    def __ge__(self,v):
        """ Handle >= comparison
        """
        if isinstance(v,DateTime):
            return self.utc_data >= v.utc_data
        elif isinstance(v,datetime.datetime):
            return self.utc_data >= v
        return NotImplemented

    def __lt__(self,v):
        """ Handle < comparison
        """
        if isinstance(v,DateTime):
            return self.utc_data < v.utc_data
        elif isinstance(v,datetime.datetime):
            return self.utc_data < v
        return NotImplemented

    def __le__(self,v):
        """ Handle <= comparison
        """
        if isinstance(v,DateTime):
            return self.utc_data <= v.utc_data
        elif isinstance(v,datetime.datetime):
            return self.utc_data <= v
        return NotImplemented

class DateTimeLocal(DateTime):
    def __init__(self,data,context=None,tz=None):
        """ Accepts either a string or an existing datetime and loads it up
        """
        if tz:
            local_tz = pytz.timezone(tz)
        elif context and context.get('tz'):
            local_tz = pytz.timezone(context['tz'])
        else:
            local_tz = pytz.utc
        super(DateTimeLocal,self).__init__(
                data,
                data_tz=local_tz,
                local_tz=local_tz,
                context=context)

class DateTimeUTC(DateTime):
    def __init__(self,data,context=None,tz=None):
        """ Accepts either a string or an existing datetime and loads it up
        """
        if tz:
            local_tz = pytz.timezone(tz)
        elif context and context.get('tz'):
            local_tz = pytz.timezone(context['tz'])
        else:
            local_tz = pytz.utc
        super(DateTimeUTC,self).__init__(
                data,
                data_tz=pytz.utc,
                local_tz=local_tz,
                context=context)



# Just for testing purposes
if __name__ == '__main__':
    context = {'tz':'America/Vancouver'}
    a = DateTimeLocal('2017-01-01 09:03:01',context=context)
    b = DateTimeLocal('2017-01-11 21:51:32',context=context)

    print(a.local_str())
    print(">",a.utc_str())

    print(b.local_str())
    print(">",b.utc_str())

    r = TimeRange(a,b)
    print("Local day difference",r.local_days())
    print("UTC day difference",r.utc_days())

    for c in r.local_chunks():
        print(c.local_str())


    d = Date('2018-02-16',context=context)
    print(d)
    print(d.local_start_time().local_str())
    print(d.utc_end_time().local_str())

    print("Trying UTC changes")
    d = DateTimeUTC(u'2018-03-01T22:05:13+00:00',context)
    print(d.utc_str())
    print(d.local_str())


    d = DateTimeUTC(u'2018-03-01T22:05:13+00:00',context)
    print(d.utc_str())
    print(d.local_str())




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['izaber']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.3',
 'PyYAML>=5.3.0,!=5.4.1,!=6.0',
 'appdirs>=1.4.4',
 'docopt>=0.6.2',
 'pytest-xdist==2.5.0',
 'python-dateutil>=2.8.1',
 'pytz>=2021.1',
 'six>=1.15.0']

extras_require = \
{'email': ['beautifulsoup4>=4.9.3', 'lxml>=4.6.2']}

setup_kwargs = {
    'name': 'izaber',
    'version': '3.1.20230803',
    'description': 'Base load point for iZaber code',
    'long_description': 'None',
    'author': 'Aki Mimoto',
    'author_email': 'aki@zaber.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.0,<4.0',
}


setup(**setup_kwargs)



import sys

app_config = {}
dependancies = {}

initializer_lookup = {}
initialization_rack = {}

def initializer(key,before=[]):
    def rack(key,f):
        initialization_rack[key] = f
        initializer_lookup[key] = f
        if before:
            for b in before:
                dependancies.setdefault(b,[])
                dependancies[b].append(key)
        return f

    return lambda f: rack(key,f)

def initialize(name=None, **kwargs):
    """ Invokes the initialization code for all the izaber.* modules
        that are hanging off of the system.
    """
    if name is None:
        name = sys.argv[0] or ''
    kwargs['name'] = name
    app_config.update(kwargs)

    # Reset load sequence if required
    if kwargs.get('force'):
        initialization_rack.update(initializer_lookup)

    # Go through load sequence handling dependancies as we go
    for key, func in dict(initializer_lookup).items():
        # If the initialization rack doesn't have the key, it means
        # we've already visited the initializer
        if not initialization_rack[key]:
            continue

        # Let's now ensure all the deps of this module are handled
        # (They need to be initialized before we can invoke this one)
        if key in dependancies:
            for b in dependancies.get(key,[]):
                result = request_initialize(b, **kwargs)
                if result:
                    kwargs = result

        # Was getting a weird error with flask where kwargs become None
        if not kwargs: kwargs = {}
        kwargs = request_initialize(key, **kwargs)
        initialization_rack[key] = None

def request_initialize(key, **kwargs):
    """ Force the initialization of another module tagged via `key`
    """

    # If the initialization rack doesn't have the key, it means
    # we've already visited the initializer
    if not initialization_rack[key]:
        return

    # Then invoke the initializer and log that we've completed
    replace_kwargs = initializer_lookup[key](**kwargs)
    initialization_rack[key] = None

    if replace_kwargs: return replace_kwargs
    return kwargs

